from bolsonaro.models.omp_forest import OmpForest, SingleOmpForest
from bolsonaro.utils import binarize_class_data

import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import OrthogonalMatchingPursuit


class OmpForestBinaryClassifier(SingleOmpForest):

    DEFAULT_SCORE_METRIC = 'indicator'

    def __init__(self, models_parameters):
        estimator = RandomForestClassifier(**models_parameters.hyperparameters,
                                           random_state=models_parameters.seed, n_jobs=-1)
        super().__init__(models_parameters, estimator)

    def _check_classes(self, y):
        assert len(set(y).difference({-1, 1})) == 0, "Classes for binary classifier must be {-1, +1}"

    def fit(self, X_forest, y_forest, X_omp, y_omp):
        self._check_classes(y_forest)
        self._check_classes(y_omp)

        return super().fit(X_forest, y_forest, X_omp, y_omp)

    def score(self, X, y, metric=DEFAULT_SCORE_METRIC):
        """
        Evaluate OMPForestClassifer on (`X`, `y`) using `metric`

        :param X:
        :param y:
        :param metric: might be "indicator"
        :return:
        """
        predictions = self.predict(X)

        if metric == 'indicator':
            evaluation = np.abs(np.mean(np.abs(np.sign(predictions) - y) - 1))
        else:
            raise ValueError("Unsupported metric '{}'.".format(metric))

        return evaluation


class OmpForestMulticlassClassifier(OmpForest):

    DEFAULT_SCORE_METRIC = 'indicator'

    def __init__(self, models_parameters):
        estimator = RandomForestClassifier(**models_parameters.hyperparameters,
                                           random_state=models_parameters.seed, n_jobs=-1)
        super().__init__(models_parameters, estimator)
        # question: peut-être initialiser les omps dans le __init__? comme pour le SingleOmpForest
        self._dct_class_omp = {}

    def fit_omp(self, atoms, objective):
        assert len(self._dct_class_omp) == 0, "fit_omp can be called only once on {}".format(self.__class__.__name__)
        possible_classes = sorted(set(objective))
        for class_label in possible_classes:
            atoms_binary = binarize_class_data(atoms, class_label, inplace=False)
            objective_binary = binarize_class_data(objective, class_label, inplace=False)
            # todo peut etre considérer que la taille de forêt est globale et donc seulement une fraction est disponible pour chaque OMP...
            omp_class = OrthogonalMatchingPursuit(
                n_nonzero_coefs=self.models_parameters.extracted_forest_size,
                fit_intercept=True, normalize=False)
            omp_class.fit(atoms_binary, objective_binary)
            self._dct_class_omp[class_label] = omp_class
        return self._dct_class_omp

    def predict(self, X):
        forest_predictions = self._base_estimator_predictions(X)

        if self._models_parameters.normalize_D:
            forest_predictions /= self._forest_norms

        label_names = []
        preds = []
        for class_label, omp_class in self._dct_class_omp.items():
            label_names.append(class_label)
            atoms_binary = binarize_class_data(forest_predictions, class_label, inplace=False)
            preds.append(self._make_omp_weighted_prediction(atoms_binary, omp_class, self._models_parameters.normalize_weights))

        # todo verifier que ce n'est pas bugué ici

        preds = np.array(preds).T
        max_preds = np.argmax(preds, axis=1)
        return np.array(label_names)[max_preds]

    def score(self, X, y, metric=DEFAULT_SCORE_METRIC):
        predictions = self.predict(X)

        if metric == 'indicator':
            evaluation = np.sum(np.ones_like(predictions)[predictions == y]) / X.shape[0]
        else:
            raise ValueError("Unsupported metric '{}'.".format(metric))

        return evaluation


if __name__ == "__main__":
    forest = RandomForestClassifier(n_estimators=10)
    X = np.random.rand(10, 5)
    y = np.random.choice([-1, +1], 10)
    forest.fit(X, y)
    print(forest.predict(np.random.rand(10, 5)))